/* Copyright 1999 PSO Systems Inc */

/*
 * This file includes VxWorks headers and exposes VxWorks facilities
 * to the rest of the system via a predefined API such as fd_get_value().
 * Each API element is supposed to be implemented in OS independent
 * way.  This file also includes facilities that are required for the
 * BSD stack to run, including runtime facilities for packet handling
 * context (other than the application context itself.  
 * This can be viewed as an example implementation for VxWorks.
 *
 * One reason for having functions in this file, as opposed to others,
 * would be that this file includes header files from the existing
 * VxWorks environment.  Other files in the port use header files from
 * BSD Unix port.  This allows a way to get access to VxWorks data
 * structures and specifics from outside this function.  All VxWorks
 * specifics are to be isolated here.
 */

#include "vxWorks.h"
#include "ioLib.h"
#include "iosLib.h"
#include "selectLib.h"
#include "wdLib.h"
#include "semLib.h"
#include "rngLib.h"
#include "taskLib.h"

#include "vxworks_only.h"
#include "common.h"

/* thread priority for the queue message thread */
#define PSO_QTPRIORITY	44		/* XXX */

extern int errno;	
extern int soo_read(), soo_write(), soo_ioctl(), soo_close();

#define	socketops	0,0,0,soo_close,soo_read,soo_write,soo_ioctl
#define SOCKET_COOKIE	"(sock)"

DEV_HDR socket_dev;
int socket_dev_id = -1;
int socket_lib_initialized = 0;

RING_ID queue_message_ring;
SEMAPHORE *queue_message_sem = 0;
int queue_message_thread = -1;

char myerror_mesg_buf[200];

extern int tick;
void init_message_targets();

void
init_rtos_port(int hzval)
{
	sysClkRateSet(hzval);
	init_message_targets();
}

unsigned int
timer_tick_rate_get()
{
	return sysClkRateGet();
}

unsigned int
timer_tick_get()
{
	return tickGet();
}

struct timeval
timeval_tick_get()
{
	static struct timeval tm;	/* XXX OK */
	unsigned int tick_val = tickGet();
	unsigned int rate = sysClkRateGet();

	tm.tv_sec = tick_val / rate;
	tm.tv_usec = (tick_val % rate) * tick;

	return tm;
}


void 
send_signal(int tid, int sig)
{
        dbg_printf("send_signal 0%d\n",sig);
	kill(tid, sig);		/* send signal to thread */
}

void 
init_socket_lib()
{
        if (socket_lib_initialized)
                return;

        socket_lib_initialized++;
	socket_dev_id = iosDrvInstall(socketops);
	if (socket_dev_id == ERROR)
		panic("init_socket_lib: can't install socket pseudo driver");
	
	socket_dev.drvNum = socket_dev_id;
	socket_dev.name = SOCKET_COOKIE;
}

int 
fd_store_value(int *fdp, int val)	
{
	/* XXX use of socket_dev.name here implies avoidance of allocating
	the string space for the device name */
	iosFdSet(*fdp, &socket_dev, socket_dev.name, val);	
	
	dbg_printf("fd_store_value: store value 0x%x for fd %d\n",
		   val, *fdp);
	return 0;
}

int 
fd_get_value(int  fdes)
{
	int val;
	
	val = iosFdValue(fdes);
	dbg_printf("fd_get_value: returning value 0x%x\n", val);
	return val;
}

void 
set_errno(int err)
{
	errno = 0x4bae0000 | err;	/* XXX cute */
}	

/*
 * TODO: add high level statistics about alloc behavior
 */
void *
memory_alloc(int sz)
{
	void *p = (void *)calloc(1, sz);

	if (!p)
		panic("memory_alloc: null");
	return p;
}

void 
memory_free(void *p)
{
	if (!p)
		panic("memory_free: null");
	cfree(p);
}

int 
fd_get_new()
{
	int fd;
	
	fd = iosFdNew(&socket_dev, socket_dev.name, 0);
	dbg_printf("fd_get_new: return fd %d\n",fd);
	return fd;
}

void
fd_free(int fd)
{
	if (iosFdValue(fd)  < 0) {
		dbg_printf("fd_free: invalid fd %d\n",fd);
		return;
	}
	dbg_printf("fd_free: freeing fd %d\n", fd);
	iosFdFree(fd);
}

/* wait 'n' micro seconds - can be called from interrupt handler so we can't
pend on anything or call taskDelay().  This should be done properly, i.e.
by using real clock counter at high res, without interrupts.  However,
we cheat here a little for now. XXX */
void
delay_micro_seconds(volatile int n)
{
        volatile int i;

        for (i = 0; i < n * 1000; i++) /* XXX ack! */
                i++;
}


/* wake one up */
void 
selwakeup_node(int selnode)	/* really a SEL_WAKEUP_NODE * type */
{
	selWakeup((SEL_WAKEUP_NODE *)selnode);
}

/* wake up everybody */
void 
selwakeup(int list, int which)	/* list is SEL_WAKEUP_LIST * type */
{
	selWakeupAll((SEL_WAKEUP_LIST *)list, (which == 0) ? SELREAD : SELWRITE);
}

void 
selrecord(SEL_WAKEUP_LIST *list, SEL_WAKEUP_NODE *selnode)
{
	if (selNodeAdd(list, selnode) < 0)
		panic("selrecord");
}

void 
selrecord_read_cancel(void *so, SEL_WAKEUP_LIST *list, 
	SEL_WAKEUP_NODE *selnode)
{
	if (selNodeDelete(list, selnode) < 0)
		panic("selrecord_read_cancel");

	/* if no more waiting clients, unset the flag */
	if (selWakeupListLen(list) == 0)
		sock_read_select_nomore(so);
}

void 
selrecord_write_cancel(void *so, SEL_WAKEUP_LIST *list, 
	SEL_WAKEUP_NODE *selnode)
{
	if (selNodeDelete(list, selnode) < 0)
		panic("selrecord_write_cancel");

	/* if no more waiting clients, unset the flag */
	if (selWakeupListLen(list) == 0)
		sock_write_select_nomore(so);
}

void *
alloc_sel(int num)
{
	SEL_WAKEUP_LIST *p;

	if (num <= 0)
		panic("alloc_sel");
	p = memory_alloc(sizeof(SEL_WAKEUP_LIST) * num);
	return (void *)p;
}

void 
init_sel(SEL_WAKEUP_LIST *p)
{
	selWakeupListInit(p);
}

void *
alloc_mutex(int num)
{
	struct mutex_type *p;

	if (num <= 0)
		panic("alloc_mutex");
	p = memory_alloc(sizeof(struct mutex_type) * num);
	return (void *) p;
}

void 
init_mutex(struct mutex_type *p)
{
	p->owner = 0;
	p->holdings = 0;	
	p->flags = 0;	
	semBInit((SEM_ID)p->sem, SEM_Q_PRIORITY, SEM_EMPTY);
}

void * 
obtain_mutex(struct mutex_type *p, int timeout, int check_id)
{
	int ret;

#ifdef MUTEX_DEBUG
	if (timeout < 0)
		panic("obtain_mutex: bad timeout val");
	if (!p)
		panic("obtain_mutex: null");
#endif
	if (check_id) {
		if (p->owner == get_current_thread()) {
			p->holdings++;
#ifdef MUTEX_DEBUG1
			if (get_current_thread() != queue_message_thread)
				printf("obtain_mutex: 0x%x already owns 0x%x\n", 
					get_current_thread(), p);
#endif
			return 0;		/* already owns it */
		}
		if (timeout) 
			ret = semTake((SEM_ID)p->sem, timeout);
		else {
			ret = semTake((SEM_ID)p->sem, WAIT_FOREVER);
			if (ret < 0)
				panic("obtain_mutex: semTake");
		}
#ifdef MUTEX_DEBUG
		if (p->owner)
			panic("obtain_mutex: owner non zero");
#endif
		p->owner = get_current_thread();
		p->holdings++;
	} else {
		if (timeout) 
			ret = semTake((SEM_ID)p->sem, timeout);
		else  {
			ret = semTake((SEM_ID)p->sem, WAIT_FOREVER);
			if (ret < 0)
				panic("obtain_mutex: semTake 2");
		}
	}
#ifdef MUTEX_DEBUG
	if (get_current_thread() != queue_message_thread)
		printf("obtain_mutex: thread 0x%x takes 0x%x\n",
			get_current_thread(), p);
#endif
	return p;	/* we got it */
}

void 
release_mutex(struct mutex_type *p, int check_id)
{
	int ret;

	if (check_id) {
		if (p) {
			if (p->owner && p->owner != get_current_thread())
				panic("release_mutex: not owner!");
			/* TODO: probably should check if owner is not set
			whether it is the first time this mutex is given.
			For now we assume it is first time we release
			this mutex if onwer is not set */
			if (p->owner) {
				p->holdings--;
				/* holdings == 0 is OK. it means we are finally
				letting go.  Whereas if holdings > 0 we are
				releasing the mutex obtained earlier in nested 
				fashion (take of mutex by the same thread
				more than once) */
				if (p->holdings < 0)
					panic("release_mutex: holdings < 0");
			}
			p->owner = 0;
			ret = semGive((SEM_ID)p->sem);
			if (ret < 0)
				panic("release_mutex: semGive");
		} else {
#ifdef MUTEX_DEBUG1
			p->holdings--;
			/* holdings can't be zero here because the last release
			of the mutex should not have p==0 */
			if (p->holdings <= 0)
				panic("release_mutex: holdings <= 0");
			if (get_current_thread() != queue_message_thread) {
				printf("release_mutex: nested 0x%x\n", p);
			}
#endif
		}
	} else {
#ifdef MUTEX_DEBUG
		if (!p)
			panic("release_mutex: null");
#endif
		ret = semGive((SEM_ID)p->sem);
		if (ret < 0)
			panic("release_mutex: semGive 2");
	}
#ifdef MUTEX_DEBUG
	if (get_current_thread() != queue_message_thread)
		printf("release_mutex: thread 0x%x releases 0x%x\n",
			get_current_thread(), p);
#endif
}

void 
mutex_info(struct mutex_type *p)
{
	SEM_ID	sp;

	sp = (SEM_ID)p->sem;
	printf("Mutex 0x%x owner 0x%x holdings %d\n", p, 
		p->owner, p->holdings);
	semShow((SEM_ID)sp, 1);
}

int 
message_available(int type)
{
	RING_ID mesg_ring;

	switch (type) {
	case MSG_TYPE_SIMPLE:
		mesg_ring = queue_message_ring;
		break;
	
	default:
		panic("get_message: wrong type");
		break;
	}

	return (!rngIsEmpty(mesg_ring));
}

void 
get_message(int type, struct mesg_arg *msg)
{
	int retval; 
	RING_ID mesg_ring;

	switch (type) {
	case MSG_TYPE_SIMPLE:
		mesg_ring = queue_message_ring;
		break;
	
	default:
		panic("get_message: wrong type");
		break;
	}

	retval = rngBufGet(mesg_ring, (char *)msg, sizeof(*msg));
	if (retval != sizeof(*msg)) {
		sprintf(myerror_mesg_buf, "get_message: type=%d",
			msg->type);
		panic(myerror_mesg_buf);
	}
}

/*
 * Thread that handles simple messages
 */
void
queue_message_loop()
{
	int ret;
	struct mesg_arg msg;
	void (*func)(void);

	for (;;) {
		ret = semTake(queue_message_sem, WAIT_FOREVER);
		if (ret < 0)
			panic("queue_message_loop: semTake");

		if (!message_available(MSG_TYPE_SIMPLE))
			continue;	/* false alarm */

		get_message(MSG_TYPE_SIMPLE, &msg);
		func = msg.ptr;
#ifdef DEBUG99
		dbg_printf("queue_message_loop: calling function 0x%x\n",(int)func);
#endif
		/* XXX need to do any further checking or mutexing here? */
		(*func)();
	}
}

/* initialize the message thread */

void
init_message_targets()
{
	queue_message_ring = rngCreate(sizeof(struct mesg_arg *) * 
					MESSAGE_RING_SIZE);
	if (!queue_message_ring)
		panic("init_message_targets");

	queue_message_sem = (SEMAPHORE *)malloc(sizeof(SEMAPHORE));
	semBInit(queue_message_sem, SEM_Q_PRIORITY, SEM_EMPTY);

	/* XXX stack probably bigger than necessary */
	queue_message_thread = taskSpawn("tPSO1", PSO_QTPRIORITY, 0, 8000, 
				(FUNCPTR)queue_message_loop, 
				0,0,0,0,0,0,0,0,0,0);
	if (queue_message_thread < 0)
		panic("can't create queue message thread");
}

void
send_message(struct mesg_arg *arg)
{
	int retval;
	RING_ID msg_ring;

	switch(arg->type) {
	case MSG_TYPE_SIMPLE:
		msg_ring = queue_message_ring;
		break;

	default:
		panic("send_message: invalid type");
		break;
	}
	retval = rngBufPut(msg_ring, (char *)arg, sizeof(*arg));
	if (retval != sizeof(*arg)) {
		sprintf(myerror_mesg_buf, "send_message: type=%d",
			arg->type);
		panic(myerror_mesg_buf);
	}
	retval = semGive(queue_message_sem);
	if (retval < 0)
		panic("send_message: semGive");
}

void
queue_timeout_function(void (*func)(void))
{
	struct mesg_arg mesg;

	mesg.type = MSG_TYPE_SIMPLE;
	mesg.ptr = func;

	send_message(&mesg);
}

void
timeout_install(func, timo, arg, ticks)
        void (*func) (void);
	struct watchdog *timo;
	void *arg;
	register int ticks;
{
	/*
	 * if it panics here it probably is calling with wrong
	 * number of arguments.  Remember our version of timeout()
	 * takes an extra argument 'timo' watchdog pointer!
	 */
	if (!func || !timo)
		panic("timeout_install");
	wdStart((WDOG_ID)timo, ticks, (FUNCPTR)queue_timeout_function, (int)func);
}

void
timeout_cancel(func, timo, arg)
        void (*func) (void);
	struct watchdog *timo;
        void *arg; 
{
	if (!timo)
		panic("timeout_cancel");
	wdCancel((WDOG_ID)timo);
}

int 
sock_select(void *so, SEL_WAKEUP_NODE *selnode)
{
	int type;
	int retval = 0;
	int revents = 0;

	type = selWakeupType(selnode);
	if (type == SELREAD) {
		retval = sock_read_select(so, selnode);
		if (retval)
			revents |=  SELREAD;
	}
	if (type == SELWRITE) {
		retval = sock_write_select(so, selnode);
		if (retval)
			revents |=  SELWRITE;
	}
	if (revents == 0) {
		if (type == SELREAD) 
			sock_read_select_indicate(so, selnode);
		if (type == SELWRITE) 
			sock_write_select_indicate(so, selnode);
	}
	if (revents)
		return 1;
	if (type != SELREAD && type != SELWRITE) {
		set_errno(EINVAL);
		return -1;
	}

	return 0;
}

int 
sock_unselect(void *so, SEL_WAKEUP_NODE *selnode)
{
	int type;

	type = selWakeupType(selnode);
	if (type == SELREAD) 
		sock_read_select_cancel(so, selnode);
	if (type == SELWRITE) 
		sock_write_select_cancel(so, selnode);
	if (type != SELREAD && type != SELWRITE) {
		set_errno(EINVAL);
		return -1;
	}

	return 0;
}

/* suspend one's own thread context.  XXX We don't provide function
to suspend other threads context yet because it can lead to bad programming.
It's safer to suspend one's own context when the thread knows that it is
right to do so.  Suspending other thread is tricky because you don't
know what resource that thread has tied up. */

void 
context_suspend()
{
	taskSuspend(taskIdSelf());
}

int 
thread_is_suspended(int thread)
{
	return taskIsSuspended(thread);
}

void
enable_thread(int thread)
{
	taskResume(thread);
}

struct watchdog *
new_timo()
{
	return (struct watchdog *)wdCreate();
}

int 
get_current_thread()
{
	/* return current thread's ID */
	return (int)taskIdCurrent;
}

/* we use spare4 field in the thread context (very VxWorks specific )
to store some contextual info */

void 
insert_context_flag(int flag)
{
	WIND_TCB *thread;

	thread = (WIND_TCB *)get_current_thread();
	/* TODO check whether flag is already set? */
	thread->spare4 |= flag;
}

void 
remove_context_flag(int flag)
{
	WIND_TCB *thread;

	thread = (WIND_TCB *)get_current_thread();
	/* TODO check whether flag is already not set? */
	thread->spare4 &= ~flag;
}

int 
context_flag_value()
{
	WIND_TCB *thread;

	thread = (WIND_TCB*)get_current_thread();
	return (thread->spare4);
}

int 
create_thread(char *name, void (*func)(void), int priority, int stacksize,
	int arg)
{
	int thread;

	thread = taskSpawn(name, priority, 0, stacksize, 
				(FUNCPTR)func, arg,0,0,0,0,0,0,0,0,0);
	return(thread);
}

void
connect_interrupt_handler(unsigned int vec, FUNCPTR func, void *arg)
{
	int retval; 

	dbg_printf("connect_interrupt_handler: vec %d func 0x%x arg 0x%x\n",
		vec, func, arg);
	retval = intConnect(vec, func, arg);
	if (retval != OK)
		panic("can't connect interrupt\n");
}

#ifdef _PSO_INTERRUPT_SUPPORT_
void
enable_interrupt_level(unsigned int level)
{
	dbg_printf("enabling irq level %d\n",level);
#ifdef PC486
	sysIntEnablePIC(level);	/* XXX */
#else
	intEnable(level);
#endif
}

void
disable_interrupt_level(unsigned int level)
{
	dbg_printf("disabling irq level %d\n",level);
#ifdef PC486
	sysIntDisablePIC(level);	/* XXX */
#else
	intDisable(level);
#endif
}

#endif /* _PSO_INTERRUPT_SUPPORT_ */

/* emulation */

int
netErrnoSet(int err)            /* XXX not sure (undocumented) */
{
        set_errno(err);
}

/*
 * FIXME
 * Some VxWorks stub routines to satisfy linker.
 * You may need to add more or less here.
 */

#define TRACE_ROUTINE(name) \
	int name() { dbg_trace("%s\n", #name); return OK; }

int 
sockLibInit()
{
        dbg_trace("sockLibInit\n");
        init_socket_lib();
        return OK;
}

int 
netLibInit()
{
	dbg_trace("netLibInit\n");
	init_bsd_compat();
	return OK;
}


TRACE_ROUTINE(netShowInit);
TRACE_ROUTINE(ifMaskSet);
TRACE_ROUTINE(ifAddrSet);
TRACE_ROUTINE(bsdSockLibInit);
TRACE_ROUTINE(sockLibAdd);
TRACE_ROUTINE(connectWithTimeout);
TRACE_ROUTINE(ifAddrGet);
TRACE_ROUTINE(routeAdd);
TRACE_ROUTINE(netJobAdd);

/* SENS usrNetwork.c seem to use the following routines */
TRACE_ROUTINE(ipAttach);
TRACE_ROUTINE(ipLibInit);
TRACE_ROUTINE(rawIpLibInit);
TRACE_ROUTINE(rawLibInit);
TRACE_ROUTINE(udpLibInit);
TRACE_ROUTINE(udpShowInit);
TRACE_ROUTINE(tcpLibInit);
TRACE_ROUTINE(tcpShowInit);
TRACE_ROUTINE(icmpLibInit);
TRACE_ROUTINE(igmpLibInit);


/* Device specific */

int 
loattach()
{
        if (ifunit("lo0"))
                return;
	dbg_trace("loattach\n");
	loopattach(0);
        ifconfig("lo0","addr",inet_addr("127.0.0.1"));

	return OK;
}

TRACE_ROUTINE(ulipInit);
TRACE_ROUTINE(elcdetach);
TRACE_ROUTINE(ultradetach);
TRACE_ROUTINE(slipInit);
TRACE_ROUTINE(eiattach);
TRACE_ROUTINE(elcattach);
TRACE_ROUTINE(ultraattach);
TRACE_ROUTINE(eexattach);
TRACE_ROUTINE(eltattach);
TRACE_ROUTINE(eneattach);
